<?php
/*******************************************************************************
*    myMVC - Model-View-Controller for PHP                                     *
*    Copyright (C) 2012 Luca Bertoncello                                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU Lesser General Public License as published  *
*    by the Free Software Foundation; version 2.1 of the License.              *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU Lesser General Public License for more details.                       *
*                                                                              *
*    You should have received a copy of the GNU Lesser General Public License  *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

/**
 * Base class for a database table's class
 * Every table must have a derived class with the check function
 * Example:
 * 
 * <code>
 * class testDB extends myMVC_BaseDB
 * {
 *   public function __construct()
 *   {
 *     parent::__construct('test',
 *                         array('id'                  => 'testDB::checkId',
 *                               'firstname'           => 'testDB::checkName',
 *                               'lastname'            => 'testDB::checkName',
 *                         ),
 *                         array('id' => 'autoincrement'),
 *                        // Unique Key: firstname+lastname
 *                         array(array('firstname', 'lastname')));
 *   }
 * 
 *   public static checkId($type, $value)
 *   {
 *     return true;
 *   }
 * ...
 * }
 * </code>
 */
class myMVC_BaseDB extends myMVC_DatabaseConnector
{
  protected $table;
  protected $pk;
  protected $fields;
  protected $uniques;
  protected $autoincrement;
  protected $values, $oldValues;
  protected $engine;
  protected $finders;
  protected $counters;

  /**
   * Constructor
   *
   * @param string $table                       The table on the database 
   * @param array $fields                       Associative array with the fields of the table and their check functions
   * @param array $pk                           Array with the PKs of the table. It can contains the key "autoincrement", if the field is an autoincrement field
   * @param array $uniques                      If not NULL, array of arrays, with the unique keys. Every entry is an array with all field for this unique key
   */
  public function __construct($table, $fields, $pk, $uniques = NULL)
  {
    global $config;

    if($config['DATABASE']['SERVER'] == '' ||
       $config['DATABASE']['USER'] == '' || 
       $config['DATABASE']['PASSWORD'] == '' ||
       $config['DATABASE']['DATABASE'] == '' ||
       ($config['DATABASE']['ENGINE'] != 'mysql' &&
        $config['DATABASE']['ENGINE'] != 'pgsql'))
    {
      print "<b>In order to use the database, you <u>must</u> define:</b><br>\n";
      print "\$config['DATABASE']['SERVER'] with the server to connect to<br>\n";
      print "\$config['DATABASE']['USER'] with the database user<br>\n";
      print "\$config['DATABASE']['PASSWORD'] with the user's password<br>\n";
      print "\$config['DATABASE']['DATABASE'] with the database name<br>\n";
      print "\$config['DATABASE']['ENGINE'] with the database type (mysql, pgsql)<br>\n";
      exit(0);
    }

    $this->finders            = array();
    $this->counters           = array();
    $engineClass              = 'myMVC_DatabaseConnector'.ucfirst($config['DATABASE']['ENGINE']);
    $this->engine             = new $engineClass();
    $this->table              = $table;
    $this->fields             = $fields;
    if(is_array($uniques))
    {
      foreach($uniques as $unique)
        if(!is_array($unique))
        {
          print "<b>The unique key is not an array of array</b>\n";
          exit(0);
        }
      $this->uniques          = $uniques;
    }
    else
      $this->uniques          = array();
    $this->pk                 = array();
    $this->autoincrement      = NULL;
    foreach(array_keys($pk) as $p)
    {
      array_push($this->pk, $p);
      if($pk[$p] == 'autoincrement')
      {
        if($this->autoincrement != NULL)
        {
          print "<b>You can just have <u>one</u> autoincrement field</b>\n";
          exit(0);
        }
        $this->autoincrement  = $p;
      }
    }
    $this->values             = array();
    foreach(array_keys($this->fields) as $f)
    {
      $this->values[$f]       = NULL;
      $this->oldValues[$f]    = NULL;
    }
  }

  /**
   * Dynamic get|set method
   * If the function does not exists (unknown get|set Key),
   * an error will be displayed
   * 
   * @param $name                               The name of the unknown called function
   * @param $args                               The arguments
   */
  public function __call($name, $args)
  {
    $found = false;
    if(strncmp($name, 'get', 3) == 0 || strncmp($name, 'set', 3) == 0)
    {
      foreach(array_keys($this->fields) as $f)
      {
        $f1 = '';
        foreach(explode('_', $f) as $p)
          $f1 .= ucfirst($p);
        if($name == 'get'.ucfirst($f1))
        {
          $found = true;
          return $this->values[$f];
        }
        if($name == 'set'.ucfirst($f1))
        {
          $found = true;
          $this->values[$f] = $args[0];
        }
      }
    }
    if(!$found)
    {
      $trace = debug_backtrace();
      printf("<br />\n<b>Fatal error</b>:  Call to undefined method %s::%s() in <b>%s</b> on line <b>%d</b><br />",
             $trace[1]['class'], $trace[1]['function'],
             $trace[1]['file'], $trace[1]['line']);
      exit(1);
    }
  }

  /**
   * Register a new finder method.
   * This can be called with the method callFinder or iterateFinder.
   * 
   * @param string $name                        The name of the finder
   * @param array $fieldsToSearch               An associative array with the fields to use for the search
   *                                            Format of the array: field => "LIKE" or field => "EQUAL"
   * @param array $fixedFields                  An associative array with the "fixed" fields for the search (default: empty)
   *                                            Format of the array: field => value
   * @throws myMVC_DatabaseException            On Problems
   */
  protected function registerFinder($name, $fieldsToSearch, $fixedFields = array())
  {
    if(array_key_exists($name, $this->finders))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Finder %s already defined", $name));
      throw new myMVC_DatabaseException($msg);
    }
    if(!is_array($fieldsToSearch))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              'Given parameter for fieldsToSearch is not an array');
      throw new myMVC_DatabaseException($msg);
    }
    if(!is_array($fixedFields))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              'Given parameter for fixedFields is not an array');
      throw new myMVC_DatabaseException($msg);
    }
    foreach(array_keys($fieldsToSearch) as $f)
      if($fieldsToSearch[$f] != 'LIKE' && $fieldsToSearch[$f] != 'EQUAL')
      {
        $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
                sprintf("Invalid search parameter %s (%s)", $f, $fieldsToSearch[$f]));
        throw new myMVC_DatabaseException($msg);
      }
    $this->finders[$name] = array('fieldsToSearch' => $fieldsToSearch, 'fixedFields' => $fixedFields);
  }

  /**
   * Calls a defined finder (see registerFinder)
   * 
   * @param string $name                        The name of the finder
   * @param array $fields                       An associative array with the fields to use for the search (field => value)
   * @param string $order                       If not NULL, field(s) to sort data
   * @param int $limit                          If not NULL, maximum of the searched data (useful for paging)
   * @param int $offset                         If not NULL, offset of the searched data (useful for paging)
   * @return array                              The result as array of database class or NULL if no record was found
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function callFinder($name, $fields, $order = NULL, $limit = NULL, $offset = NULL)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    $result = $db->callIterator($name, $fields, $order, $limit, $offset);
    if($result === NULL)
      return NULL;

    $ret = array();
    while($row = $db->engine->getValueArrayFromResult($result))
    {
      $rowClass = new $calledClass();
      $rowClass->saveValues($row);
      $ret[] = $rowClass;
    }
    $db->engine->freeResult($result);
    return $ret;
  }

  /**
   * Iterate throws a defined finder (see registerFinder)
   * 
   * @param string $name                        The name of the finder
   * @param array $fields                       An associative array with the fields to use for the search (field => value)
   * @param string $order                       If not NULL, field(s) to sort data
   * @param int $limit                          If not NULL, maximum of the searched data (useful for paging)
   * @param int $offset                         If not NULL, offset of the searched data (useful for paging)
   * @return resource                           The result as database resource or NULL if no record was found
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function callIterator($name, $fields, $order = NULL, $limit = NULL, $offset = NULL)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    if(!array_key_exists($name, $db->finders))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Finder %s not defined", $name));
      throw new myMVC_DatabaseException($msg);
    }
    if(!is_array($fields))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              'Given parameter for fields is not an array');
      throw new myMVC_DatabaseException($msg);
    }

    $fieldsToSearch = $db->finders[$name]['fieldsToSearch'];
    $fixedFields = $db->finders[$name]['fixedFields'];
    $searchFields = array();
    $useLike = array();
    foreach(array_keys($fixedFields) as $f)
      $searchFields[$f] = $fixedFields[$f];
    foreach(array_keys($fieldsToSearch) as $f)
      if($fieldsToSearch[$f] == 'LIKE')
        $useLike[$f] = $f;
    $c = 0;
    foreach(array_keys($fields) as $f)
    {
      $found = false;
      foreach(array_keys($fieldsToSearch) as $p)
        if($f == $p)
        {
          $found = true;
          $c++;
          $searchFields[$f] = $fields[$f];
        }
      if(!$found)
      {
        $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
                sprintf("%s is not part of the search field for finder %s::%s", $f, $calledClass, $name));
        throw new myMVC_DatabaseException($msg);
      }
    }
    if(count($fieldsToSearch) != $c)
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Not all fields of the finder %s::%s are specified", $calledClass, $name));
      throw new myMVC_DatabaseException($msg);
    }

    $result = $db->findByParameter($searchFields, $useLike, $order, $limit, $offset);
    if($result === NULL)
      return NULL;
    return $result;
  }

  /**
   * Register a new counter method.
   * This can be called with the method callCounter.
   * 
   * @param string $name                        The name of the counter
   * @param array $fieldsToSearch               An associative array with the fields to use for the search
   *                                            Format of the array: field => "LIKE" or field => "EQUAL"
   * @param array $fixedFields                  An associative array with the "fixed" fields for the search (default: empty)
   *                                            Format of the array: field => value
   * @throws myMVC_DatabaseException            On Problems
   */
  protected function registerCounter($name, $fieldsToSearch, $fixedFields = array())
  {
    if(array_key_exists($name, $this->counters))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Counter %s already defined", $name));
      throw new myMVC_DatabaseException($msg);
    }
    if(!is_array($fieldsToSearch))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              'Given parameter for fieldsToSearch is not an array');
      throw new myMVC_DatabaseException($msg);
    }
    if(!is_array($fixedFields))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              'Given parameter for fixedFields is not an array');
      throw new myMVC_DatabaseException($msg);
    }
    foreach(array_keys($fieldsToSearch) as $f)
      if($fieldsToSearch[$f] != 'LIKE' && $fieldsToSearch[$f] != 'EQUAL')
      {
        $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
                sprintf("Invalid search parameter %s (%s)", $f, $fieldsToSearch[$f]));
        throw new myMVC_DatabaseException($msg);
      }
    $this->counters[$name] = array('fieldsToSearch' => $fieldsToSearch, 'fixedFields' => $fixedFields);
  }

  /**
   * Calls a defined counter (see registerCounter)
   * 
   * @param string $name                        The name of the counter
   * @param array $fields                       An associative array with the fields to use for the search (field => value)
   * @return int                                The count of the searched records
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function callCounter($name, $fields)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    if(!array_key_exists($name, $db->counters))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Counter %s not defined", $name));
      throw new myMVC_DatabaseException($msg);
    }
    if(!is_array($fields))
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              'Given parameter for fields is not an array');
      throw new myMVC_DatabaseException($msg);
    }

    $fieldsToSearch = $db->counters[$name]['fieldsToSearch'];
    $fixedFields = $db->counters[$name]['fixedFields'];
    $searchFields = array();
    $useLike = array();
    foreach(array_keys($fixedFields) as $f)
      $searchFields[$f] = $fixedFields[$f];
    foreach(array_keys($fieldsToSearch) as $f)
      if($fieldsToSearch[$f] == 'LIKE')
        $useLike[$f] = $f;
    $c = 0;
    foreach(array_keys($fields) as $f)
    {
      $found = false;
      foreach(array_keys($fieldsToSearch) as $p)
        if($f == $p)
        {
          $found = true;
          $c++;
          $searchFields[$f] = $fields[$f];
        }
      if(!$found)
      {
        $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
                sprintf("%s is not part of the search field for counter %s::%s", $f, $calledClass, $name));
        throw new myMVC_DatabaseException($msg);
      }
    }
    if(count($fieldsToSearch) != $c)
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Not all fields of the counter %s::%s are specified", $calledClass, $name));
      throw new myMVC_DatabaseException($msg);
    }

    return $db->getCountByParameter($searchFields, $useLike);
  }

  /**
   * Saves the values of the record
   * 
   * @param array $values                       Associative array with the values from the database
   */
  protected function saveValues($values)
  {
    $this->oldValues = $this->values = $values;
  }

  /**
   * Returns all records
   * 
   * @param string $order                       If not NULL, field(s) to sort data
   * @param int $limit                          If not NULL, maximum of the searched data (useful for paging)
   * @param int $offset                         If not NULL, offset of the searched data (useful for paging)
   * @return array                              The result as array of database class or NULL if no record was found
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function findAll($order = NULL, $limit = NULL, $offset = NULL)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    $result = $db->iterateAll($order, $limit, $offset);
    if($result === NULL)
      return NULL;

    $ret = array();
    while($row = $db->engine->getValueArrayFromResult($result))
    {
      $rowClass = new $calledClass();
      $rowClass->saveValues($row);
      $ret[] = $rowClass;
    }
    $db->engine->freeResult($result);
    return $ret;
  }

  /**
   * Iterate all records
   * 
   * @param string $order                       If not NULL, field(s) to sort data
   * @param int $limit                          If not NULL, maximum of the searched data (useful for paging)
   * @param int $offset                         If not NULL, offset of the searched data (useful for paging)
   * @return resource                           The result as database resource or NULL if no record was found
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function iterateAll($order = NULL, $limit = NULL, $offset = NULL)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    $result = $db->findByParameter(array(), array(), $order, $limit, $offset);
    if($result === NULL)
      return NULL;
    return $result;
  }

  /**
   * Returns the count of all records
   * 
   * @return int                                The count of all records
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function getCountAll()
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    return $db->getCountByParameter(array());
  }

  /**
   * Get a record using the primary key of the table
   * 
   * @param array $pk                           The primary key as associative array (field => value)
   * @return DB-Class                           The result as database class or NULL if no record was found
   * @throws myMVC_DatabaseException            On Problems
   */
  public static function findByPK($pk)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();

    $c = 0;
    foreach(array_keys($pk) as $p)
    {
      $found = false;
      foreach($db->pk as $p1)
        if($p == $p1)
        {
          $found = true;
          $c++;
        }
      if(!$found)
      {
        $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
                sprintf("%s is not part of the primary key for class %s", $p, $calledClass));
        throw new myMVC_DatabaseException($msg);
      }
    }
    if(count($db->pk) != $c)
    {
      $msg = myMVC_Logger::doLog(__METHOD__, __LINE__,
              sprintf("Not all fields of the primary key for class %s are specified", $calledClass));
      throw new myMVC_DatabaseException($msg);
    }
    $result = $db->findByParameter($pk);
    if($result === NULL)
      return NULL;

    $db->saveValues($db->engine->getValueArrayFromResult($result));
    $db->engine->freeResult($result);
    return $db;
  }

  /**
   * Creates the WHERE-statement for the SQL query, using parameters
   * 
   * @param array $param                        The parameters, as associative array field => value
   * @param array $useLike                      Associative array with the fields that should be search with a LIKE (default: empty)
   * @returns string                            The WHERE-Statement
   */
  protected function createWhere($param, $useLike = array())
  {
    if(count($param) == 0)
      return '';

    $first = true;
    $where = 'WHERE';
    foreach(array_keys($param) as $key)
    {
      if($first)
        $first = false;
      else
        $where .= ' AND';
      if(is_array($param[$key]))
      {
        $where .= ' `'.$key.'` IN (';
        $first = true;
        foreach($param[$key] as $v)
        {
          if($first)
            $first = false;
          else
            $where .= ', ';
          $where .= $this->engine->quoteSqlString($v);
        }
        $where .= ')';
      }
      else if($param[$key] === NULL)
        $where .= sprintf(" `%s` IS NULL", $key);
      else if(!is_numeric($param[$key]))
        $where .= sprintf(" %s %s %s",
                      ($useLike[$key] ? 'LOWER(`'.$key.'`)' : '`'.$key.'`'),
                      ($useLike[$key] ? 'LIKE' : '='),
                      ($useLike[$key] ? strtolower($this->engine->quoteSqlStringForLike($param[$key])) : $this->engine->quoteSqlString($param[$key])));
      else
        $where .= sprintf(" %s = %s", $key, $param[$key]);
    }
    return $where;
  }

  /**
   * Get a record from the database, using parameters
   * 
   * @param array $param                        The parameters, as associative array field => value
   * @param array $useLike                      Associative array with the fields that should be search with a LIKE (default: empty)
   * @param string $order                       If not NULL, field(s) to sort data
   * @param int $limit                          If not NULL, maximum of the searched data (useful for paging)
   * @param int $offset                         If not NULL, offset of the searched data (useful for paging)
   * @return resource                           The result as database resource or NULL if no data was found
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On generic problem
   */
  protected function findByParameter($param, $useLike = array(), $order = NULL, $limit = NULL, $offset = NULL)
  {
    $this->checkFields($param, 'search');

    $fields = '';
    $first = true;
    foreach(array_keys($this->fields) as $k)
    {
      if($first)
        $first = false;
      else
        $fields .= ', ';
      $fields .= '`'.$k.'`';
    }

    $sql = sprintf("SELECT %s FROM %s %s", $fields, $this->table, $this->createWhere($param, $useLike));
    if($order !== NULL && trim($order) != '')
      $sql .= ' ORDER BY '.$order;
    if($limit !== NULL && trim($limit) != '')
      $sql .= ' LIMIT '.$limit;
    if($offset !== NULL && trim($offset) != '')
      $sql .= ' OFFSET '.$offset;

    return $this->findBySQLQuery($sql);
  }

  /**
   * Get the count of the records, using parameters
   * 
   * @param array $param                        The parameters, as associative array field => value
   * @param array $useLike                      Associative array with the fields that should be search with a LIKE (default: empty)
   * @return int                                The count of the record
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On generic problem
   */
  protected function getCountByParameter($param, $useLike = array())
  {
    $this->checkFields($param, 'search');

    $sql = sprintf("SELECT COUNT(*) AS c FROM %s %s", $this->table, $this->createWhere($param, $useLike));
    $result = $this->findBySQLQuery($sql);
    $row = $this->engine->getValueArrayFromResult($result);
    $this->engine->freeResult($result);
    return $row['c'];
  }

  /**
   * Sends to the database engine an SQL query and get the result
   * 
   * @param string $sql                         The SQL query
   * @return resource                           The result as database resource or NULL if no data was found
   * @throws myMVC_DatabaseException            On generic problem
   */
  protected function findBySQLQuery($sql)
  {
    if(($ret = $this->engine->executeSelect($sql)) !== false)
    {
      if($this->engine->getRowsNum($ret) == 0)
        return NULL;
      return $ret;
    }
    else
      throw new myMVC_DatabaseException(sprintf("Error on SQL query. Errno: %d - %s", $this->engine->errno(), $this->engine->error()));
  }

  /**
   * Checks if all the given data are valid
   * 
   * @param array $param                        The data to check
   * @param string $type                        Check type (save/search)
   * @throws myMVC_InvalidFieldException        On invalid field
   */
  protected function checkFields($param, $type)
  {
    foreach(array_keys($param) as $key)
    {
      if(array_key_exists($key, $this->fields))
        $checkFunc = $this->fields[$key];
      if(!is_array($param[$key]))
        if(!call_user_func(explode('::', $checkFunc), $param[$key], $type))
          throw new myMVC_InvalidFieldException(sprintf('Invalid field %s (%s)', $key, $param[$key] !== NULL ? $param[$key] : 'NULL'));
    }
  }

  /**
   * Check whether the current record satisfies all unique keys
   * 
   * @return bool                               true if all unique keys are satisfied, false otherwise
   */
  public function isRecordUnique()
  {
    $conds = '';
    foreach($this->uniques as $unique)
    {
      $key = array();
      foreach($unique as $u)
        $key[$u] = $this->values[$u];
      $w = '('.str_replace('WHERE ', '', $this->createWhere($key, false)).')';
      if($conds == '')
        $conds = $w;
      else
        $conds .= ' OR '.$w;
    }
    $pk = '';
    foreach($this->pk as $key)
    {
      if($pk == '')
        $pk = sprintf("`%s` != %s", $key, $this->engine->quoteSqlString($this->values[$key]));
      else
        $pk .= sprintf(" AND `%s` != %s", $key, $this->engine->quoteSqlString($this->value[$key]));
    }
    if($conds != '')
    {
      if($pk != '')
        $sql = sprintf("SELECT COUNT(*) AS c FROM %s WHERE %s AND (%s)", $this->table, $pk, $conds);
      else
        $sql = sprintf("SELECT COUNT(*) AS c FROM %s WHERE %s", $this->table, $conds);
      $result = $this->findBySQLQuery($sql);
      $row = $this->engine->getValueArrayFromResult($result);
      $this->engine->freeResult($result);
      if($row['c'] != 0)
        return false;
    }
    return true;
  }

  /**
   * Checks if the unique keys are satisfied
   * 
   * @throws myMVC_InvalidFieldException        On unsatisfied unique key
   */
  protected function checkUniques()
  {
    if(!$this->isRecordUnique())
      throw new myMVC_InvalidFieldException('Duplicated record. At least one unique key was NOT satisfied');
  }

  /**
   * Save the data. It checks the PK: if the record does not exists, create a new record.
   * Otherwise it updates it.
   * 
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On generic problem
   */
  public function save()
  {
    $this->checkFields($this->values, 'save');
    $this->checkUniques();

    $keys = $values = $keyValues = '';
    $first = $firstUpd = true;
    foreach(array_keys($this->values) as $key)
    {
      if($this->values[$key] != $this->oldValues[$key])
      {
        if($firstUpd)
          $firstUpd = false;
        else
          $keyValues .= ', ';
        if($this->values[$key] !== NULL)
          $keyValues .= '`'.$key.'` = '.$this->engine->quoteSqlString($this->values[$key]);
        else
          $keyValues .= '`'.$key.'` = NULL';
      }
      if($this->values[$key] !== NULL)
      {
        if($first)
          $first = false;
        else
        {
          $keys .= ', ';
          $values .= ', ';
        }
        $keys .= '`'.$key.'`';
        $values .= $this->engine->quoteSqlString($this->values[$key]);
      }
      $this->oldValues[$key] = $this->values[$key];
    }
    if($keyValues != '')
    { // $keyValues will always set if a save is needed
      $sql = sprintf("INSERT INTO %s (%s) VALUES (%s) ON DUPLICATE KEY UPDATE %s", $this->table, $keys, $values, $keyValues);
      $ret = $this->engine->execute($sql);
      if($ret === false)
        throw new myMVC_DatabaseException('Unable to save into '.$this->table);
      if($this->autoincrement !== NULL)
        $this->values[$this->autoincrement] = $ret;
    }
  }

  /**
   * Deletes the record, using the primary key
   * 
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On generic problem
   */
  public function delete()
  {
    $param = array();
    foreach($this->pk as $key)
      $param[$key] = $this->values[$key];
    $ret = $this->engine->execute(sprintf("DELETE FROM %s %s", $this->table, $this->createWhere($param)));
    if($ret === false)
      throw new myMVC_DatabaseException('Unable to DELETE '.$this->table);
  }

  /**
   * Deletes all records that match the given parameters
   * 
   * @param array $param                        The parameters, as associative array field => value
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On generic problem
   */
  public static function deleteAll($param)
  {
    $calledClass = get_called_class();
    $db = new $calledClass();
    $ret = $db->engine->execute(sprintf("DELETE FROM %s %s", $db->table, $db->createWhere($param)));
    if($ret === false)
      throw new myMVC_DatabaseException('Unable to DELETE '.$db->table);
  }

  /**
   * Export data as SQL queries, using parameters
   * Exportiert Datensätze als SQL anhand der eingegebenen Parametern
   *
   * @param array $param                        The parameters, as associative array field => value
   * @param array $useLike                      Associative array with the fields that should be search with a LIKE (default: empty)
   * @param string $order                       If not NULL, field(s) to sort data
   * @param int $limit                          If not NULL, maximum of the searched data (useful for paging)
   * @param int $offset                         If not NULL, offset of the searched data (useful for paging)
   * @param boolean $exportToBrowser            If true (default) data will be exported to the browser
   *                                            If false, data will be exported in the return string
   * @param int $separateAfter                  After $separateAfter lines (default 50) a new SQL header (INSERT INTO...) will be added
   * @return string                             The export as SQL queries or NULL if no data is available
   * @throws myMVC_InvalidFieldException        On invalid field
   * @throws myMVC_DatabaseException            On generic problem
   */
  public function sqlExportByParameter($param, $useLike = array(), $order = NULL, $limit = NULL, $offset = NULL, $exportToBrowser = true, $separateAfter = 50)
  {
    $exportedSQL = '';
    $insertSql = 'INSERT INTO '.$this->table.' (';

    $first = true;
    foreach(array_keys($this->fields) as $k)
    {
      if($first)
        $first = false;
      else
        $insertSql .= ', ';
      $insertSql .= '`'.$k.'`';
    }
    $insertSql .= ") VALUES\n";

    $res = $this->findByParameter($param, $useLike, $order, $limit, $offset);
    $firstLine = true;
    $nLine = 0;
    $first = true;
    while($row = $this->engine->getValueArrayFromResult($res))
    {
      if($firstLine)
      {
        if($first)
          $first = false;
        else
        {
          if($exportToBrowser)
            print ";\n";
          else
            $exportedSQL .= ";\n";
        }
        $firstLine = false;
        if($exportToBrowser)
          print $insertSql;
        else
          $exportedSQL .= $insertSql;
      }
      else
      {
        if($exportToBrowser)
          print ",\n";
        else
          $exportedSQL .= ",\n";
      }
      $sqlLine = '(';
      foreach(array_keys($row) as $rk)
      {
        if($sqlLine != '(')
          $sqlLine .= ', ';
        $sqlLine .= $this->engine->quoteSqlString($row[$rk]);
      }
      $sqlLine .= ')';
      $nLine++;
      if($nLine == $separateAfter)
      {
        $firstLine = true;
        $nLine = 0;
      }
      if($exportToBrowser)
        print $sqlLine;
      else
        $exportedSQL .= $sqlLine;
    }
    $this->engine->freeResult($res);
    if($exportToBrowser)
      print ";\n";
    else
      $exportedSQL .= ";\n";

    return $exportedSQL;
  }

  /**
   * Returns the current date and time, formatted as for insert in the database
   * 
   * @return string                             The current date and time
   */
  public function now()
  {
    return date('Y-m-d H:i:s');
  }

  /**
   * Returns the current date and time, formatted as in the current locale
   * 
   * @return string                             The current date and time
   */
  public function nowFormatted()
  {
    global $config;

    return date($config['DATETIMEFORMAT']);
  }
}
?>
